# -*- mode: python; coding: utf-8 -*-
# Copyright 2013-2020 Chris Beaumont and the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""Computations for the TOAST projection scheme and tile pyramid format.

TODO this all needs to be ported to modern Toasty infrastructure and
wwt_data_formats.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
generate_tiles
minmax_tile_filter
nxy_tile_filter
sample_layer
Tile
toast_tile_area
'''.split()

from collections import defaultdict, namedtuple
import os
import logging
import numpy as np
from tqdm import tqdm

from ._libtoasty import subsample, mid
from .image import Image
from .norm import normalize
from .pyramid import Pos, depth2tiles, is_subtile, pos_parent, tiles_at_depth

level1 = [
    [np.radians(c) for c in row]
    for row in [
        [(0, -90), (90, 0), (0, 90), (180, 0)],
        [(90, 0), (0, -90), (0, 0), (0, 90)],
        [(0, 90), (0, 0), (0, -90), (270, 0)],
        [(180, 0), (0, 90), (270, 0), (0, -90)],
    ]
]

Tile = namedtuple('Tile', 'pos corners increasing')


def _arclength(lat1, lon1, lat2, lon2):
    """Compute the length of an arc along the great circle defined by spherical
    latitude and longitude coordinates. Inputs and return value are all in
    radians.

    """
    c = np.sin(lat1) * np.sin(lat2) + np.cos(lon1 - lon2) * np.cos(lat1) * np.cos(lat2)
    return np.arccos(c)


def _spherical_triangle_area(lat1, lon1, lat2, lon2, lat3, lon3):
    """Compute the area of the specified spherical triangle in steradians. Inputs
    are in radians. From https://math.stackexchange.com/a/66731 . My initial
    implementation used unit vectors on the sphere instead of latitudes and
    longitudes; there might be a faster way to do things in lat/lon land.

    """
    c = _arclength(lat1, lon1, lat2, lon2)
    a = _arclength(lat2, lon2, lat3, lon3)
    b = _arclength(lat3, lon3, lat1, lon1)
    s = 0.5 * (a + b + c)
    tane4 = np.sqrt(np.tan(0.5 * s) * np.tan(0.5 * (s - a)) * np.tan(0.5 * (s - b)) * np.tan(0.5 * (s - c)))
    e = 4 * np.arctan(tane4)
    return e


def toast_tile_area(tile):
    """Calculate the area of a TOAST tile in steradians.

    Parameters
    ----------
    tile : :class:`Tile`
        A TOAST tile.

    Returns
    -------
    The area of the tile in steradians.

    Notes
    -----
    This computation is not very fast.

    """
    ul, ur, lr, ll = tile.corners

    if tile.increasing:
        a1 = _spherical_triangle_area(ul[1], ul[0], ur[1], ur[0], ll[1], ll[0])
        a2 = _spherical_triangle_area(ur[1], ur[0], lr[1], lr[0], ll[1], ll[0])
    else:
        a1 = _spherical_triangle_area(ul[1], ul[0], ur[1], ur[0], lr[1], lr[0])
        a2 = _spherical_triangle_area(ul[1], ul[0], ll[1], ll[0], lr[1], lr[0])

    return a1 + a2


def _minmax(arr):
    return min(arr), max(arr)


def minmax_tile_filter(ra_range, dec_range):
    """Returns the tile_filter function based on a ra/dec range.

    Parameters
    ----------
    ra_range, dec_range: (array)
        The ra and dec ranges to be toasted (in the form [min,max]).
    """

    def is_overlap(tile):
        c = tile[1]

        minRa,maxRa = _minmax([(x[0] + 2*np.pi) if x[0] < 0 else x[0] for x in [y for y in c if np.abs(y[1]) !=  np.pi/2]])
        minDec,maxDec = _minmax([x[1] for x in c])
        if (dec_range[0] > maxDec) or (dec_range[1] < minDec): # tile is not within dec range
            return False
        if (maxRa - minRa) > np.pi: # tile croses circle boundary
            if (ra_range[0] < maxRa) and (ra_range[1] > minRa): # tile is not within ra range
                return False
        else:
            if (ra_range[0] > maxRa) or (ra_range[1] < minRa): # tile is not within ra range
                return False

        return True

    return is_overlap


def nxy_tile_filter(layer,tx,ty):
    """Returns the tile_filter function based on a given super-tile.

    Parameters
    ----------
    layer,tx,ty: (int)
        Layer and x,y coordinates, for a tile that will serve at the "super-tile"
        such that all subtiles will be toasted/merged.
    """


    regionLoc = Pos(n=layer,x=tx,y=ty)

    def is_overlap(tile):
        tileLoc = tile[0]

        if tileLoc.n > regionLoc.n:
            return True

        return is_subtile(regionLoc,tileLoc)

    return is_overlap


def _postfix_corner(tile, depth, bottom_only, tile_filter):
    """
    Yield subtiles of a given tile, in postfix (deepest-first) order.

    Parameters
    ----------
    tile : Tile
        Parameters of the current tile.
    depth : int
        The depth to descend to.
    bottom_only : bool
        If True, only yield tiles at max_depth.
    tile_filter : callable
        A function with signature ``tile_filter(tile) -> bool`` that determines
        which tiles will be yielded; tiles for which it returns ``False`` will
        be skipped.

    """
    n = tile[0].n
    if n > depth:
        return

    if not tile_filter(tile):
        return

    for child in _div4(tile):
        for item in _postfix_corner(child, depth, bottom_only, tile_filter):
            yield item

    if n == depth or not bottom_only:
        yield tile


def _div4(tile):
    """Return the four child tiles of an input tile."""
    n, x, y = tile.pos.n, tile.pos.x, tile.pos.y
    ul, ur, lr, ll = tile.corners
    increasing = tile.increasing

    to = mid(ul, ur)
    ri = mid(ur, lr)
    bo = mid(lr, ll)
    le = mid(ll, ul)
    ce = mid(ll, ur) if increasing else mid(ul, lr)

    n += 1
    x *= 2
    y *= 2

    return [
        Tile(Pos(n=n, x=x,     y=y    ), (ul, to, ce, le), increasing),
        Tile(Pos(n=n, x=x + 1, y=y    ), (to, ur, ri, ce), increasing),
        Tile(Pos(n=n, x=x,     y=y + 1), (le, ce, bo, ll), increasing),
        Tile(Pos(n=n, x=x + 1, y=y + 1), (ce, ri, lr, bo), increasing),
    ]


def generate_tiles(depth, bottom_only=True, tile_filter=None):
    """Generate a pyramid of TOAST tiles in deepest-first order.

    Parameters
    ----------
    depth : int
        The tile depth to recurse to.
    bottom_only : bool
        If True, then only the lowest tiles will be yielded.
    tile_filter : callable or None (the default)
        If not None, a filter function applied to the process;
        only tiles for which ``tile_filter(tile)`` returns True
        will be yielded

    Yields
    ------
    tile : Tile
        An individual tile to process. Tiles are yield deepest-first.

    The ``n = 0`` depth is not included.

    """
    if tile_filter is None:
        tile_filter = lambda t: True

    todo = [
        Tile(Pos(n=1, x=0, y=0), level1[0], True),
        Tile(Pos(n=1, x=1, y=0), level1[1], False),
        Tile(Pos(n=1, x=1, y=1), level1[2], True),
        Tile(Pos(n=1, x=0, y=1), level1[3], False),
    ]

    for t in todo:
        for item in _postfix_corner(t, depth, bottom_only, tile_filter):
            yield item


def sample_layer(pio, mode, sampler, depth, cli_progress=False):
    """Generate a layer of the TOAST tile pyramid through direct sampling.

    Parameters
    ----------
    pio : :class:`toasty.pyramid.PyramidIO`
        A :class:`~toasty.pyramid.PyramidIO` instance to manage the I/O with
        the tiles in the tile pyramid.
    mode : :class:`toasty.image.ImageMode`
        The image mode of this data source.
    sampler : callable
        The sampler callable that will produce data for tiling.
    depth : int
        The depth of the layer of the TOAST tile pyramid to generate. The
        number of tiles in each layer is ``4**depth``. Each tile is 256×256
        TOAST pixels, so the resolution of the pixelization at which the
        data will be sampled is a refinement level of ``2**(depth + 8)``.
    cli_progress : optional boolean, defaults False
        If true, a progress bar will be printed to the terminal using tqdm.

    """
    with tqdm(total=tiles_at_depth(depth), disable=not cli_progress) as progress:
        for tile in generate_tiles(depth, bottom_only=True):
            lon, lat = subsample(
                tile.corners[0],
                tile.corners[1],
                tile.corners[2],
                tile.corners[3],
                256,
                tile.increasing,
            )
            sampled_data = sampler(lon, lat)
            pio.write_toasty_image(tile.pos, Image.from_array(mode, sampled_data))
            progress.update(1)

    if cli_progress:
        print()



# -*- mode: python; coding: utf-8 -*-
# This file and its tests is adapted from Glue (http://glueviz.org).
#
# Copyright (c) 2013, Glue Developers
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
# * Neither the name of the Glue project nor the names of its contributors may be
#   used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, division, print_function

import numpy as np


def norm(x, vmin, vmax):
    """
    Linearly scale data between [vmin, vmax] to [0, 1]. Clip outliers
    """
    result = (x - 1.0 * vmin)
    result = np.divide(result, vmax - vmin, out=result)
    result = np.clip(result, 0, 1, out=result)
    return result


def cscale(x, bias, contrast):
    """
    Apply bias and contrast scaling. Overwrite input

    Parameters
    ----------
    x : array
      Values between 0 and 1
    bias : float
    contrast : float

    Returns
    -------
    The input x, scaled inplace
    """
    x = np.subtract(x, bias, out=x)
    x = np.multiply(x, contrast, out=x)
    x = np.add(x, 0.5, out=x)
    x = np.clip(x, 0, 1, out=x)
    return x


def linear_warp(x, vmin, vmax, bias, contrast):
    return cscale(norm(x, vmin, vmax), bias, contrast)


def log_warp(x, vmin, vmax, bias, contrast, exp=1000.0):
    black = x < vmin
    x = norm(x, vmin, vmax)
    x = np.multiply(exp, x, out=x)
    # sidestep numpy bug that masks log(1)
    # when out is provided
    x = np.add(x, 1.001, out=x)
    x = np.log(x, out=x)
    x = np.divide(x, np.log(exp + 1.0), out=x)
    x = cscale(x, bias, contrast)
    return x


def pow_warp(x, vmin, vmax, bias, contrast, exp=1000.0):
    x = norm(x, vmin, vmax)
    x = np.power(exp, x, out=x)
    x = np.subtract(x, 1, out=x)
    x = np.divide(x, exp - 1)
    x = cscale(x, bias, contrast)
    return x


def sqrt_warp(x, vmin, vmax, bias, contrast):
    x = norm(x, vmin, vmax)
    x = np.sqrt(x, out=x)
    x = cscale(x, bias, contrast)
    return x


def squared_warp(x, vmin, vmax, bias, contrast):
    x = norm(x, vmin, vmax)
    x = np.power(x, 2, out=x)
    x = cscale(x, bias, contrast)
    return x


def asinh_warp(x, vmin, vmax, bias, contrast):
    x = norm(x, vmin, vmax)
    x = np.divide(np.arcsinh(np.multiply(x, 10, out=x), out=x), 3, out=x)
    x = cscale(x, bias, contrast)
    return x


def sinh_warp(x, vmin, vmax, bias, contrast):
    x = norm(x, vmin, vmax)
    x = np.divide(np.sinh(np.multiply(x, 3, out=x), out=x), 10, out=x)
    x = cscale(x, bias, contrast)
    return x

warpers = dict(linear=linear_warp,
               log=log_warp,
               sqrt=sqrt_warp,
               power=pow_warp,
               squared=squared_warp,
               arcsinh=asinh_warp,
               sinh=sinh_warp)

def normalize(value, vmin, vmax, bias=.5, contrast=1, stretch='linear'):
    """Normalize data, by scaling between vmin and vmax, with given bias, contrast, and stretch.

    Parameters
    ----------
    value: array
      values to be normalized
    vmin,vmax: float
      minimum and maximum values for the normalized output
    bias: float (default 0.5)
      bias scaling to be applied during normalization 
    contrast: float (default 1)
      contrast scaling to be applied during normalization
    stretch: string (default 'linear')
      stretch to apply to the data
      valid values are: linear,log,sqrt,power,squared,arcsinh,sinh

    Returns
    -------
    The normalized data array
    """
    
    inverted = vmax <= vmin

    hi, lo = max(vmin, vmax), min(vmin, vmax)

    if stretch not in warpers:
        raise ValueError("Invalid stretch option %s. Valid options are %s" %
                         (stretch, warpers.keys()))

    warp = warpers[stretch]
    result = warp(value, lo, hi, bias, contrast)

    if inverted:
        result = np.subtract(1, result, out=result)

    result = np.multiply(255, result, out=result)
    result = np.clip(result, 0, 255, out=result)
    return result.astype(np.uint8)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""General tools for merging and downsampling tiles

The Merger Protocol
-------------------

A “merger” is a callable that takes four input tiles and downsamples them into
a smaller tile. Its prototype is ``merger(big) -> small``, where *big* is a
Numpy array of at least 2 dimensions whose first two axes are both 512
elements in size. The return value *small* should have the same number of
dimensions as the input, two initial axes of size 256, and remaining axes the
same size as the input.

To efficiently vectorize two-by-two downsampling, a useful trick is to reshape
the ``(512, 512)`` input tile into a shape ``(256, 2, 256, 2)``. You can then
use functions like ``np.mean()`` with an argument ``axes=(1, 3)`` to vectorize
the operation over sets of four adjacent pixels.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
averaging_merger
cascade_images
'''.split()

import numpy as np
from tqdm import tqdm

from . import pyramid
from .image import Image


def averaging_merger(data):
    """A merger function that averages quartets of pixels.

    Parameters
    ----------
    data : array
        See the Merger Protocol specification.

    Returns
    -------
    A downsampled array. See the Merger Protocol specification.

    """
    s = (data.shape[0] // 2, 2, data.shape[1] // 2, 2) + data.shape[2:]
    return np.nanmean(data.reshape(s), axis=(1, 3)).astype(data.dtype)


def cascade_images(pio, mode, start, merger, cli_progress=False):
    """Downsample image tiles all the way to the top of the pyramid.

    This function will walk the tiles in the tile pyramid, merging child tile
    images and writing new tile images at shallower levels of the pyramid.

    Parameters
    ----------
    pio : :class:`toasty.pyramid.PyramidIO`
        An object managing I/O on the tiles in the pyramid.
    mode : :class:`toasty.image.ImageMode`
        The image mode (i.e., RGB or scientific) to process.
    start : nonnegative integer
        The depth at which to start the cascade process. It is assumed that
        the tiles *at this depth* are already populated by some other means.
        This function will create new tiles at shallower depths.
    merger : a merger function
        The method used to create a parent tile from its child tiles. This
        is a callable that follows the Merger Protocol.
    cli_progress : optional boolean, defaults False
        If true, a progress bar will be printed to the terminal using tqdm.

    """
    buf = None
    SLICES = [
        (slice(None, 256), slice(None, 256)),
        (slice(None, 256), slice(256, None)),
        (slice(256, None), slice(None, 256)),
        (slice(256, None), slice(256, None)),
    ]

    with tqdm(total=pyramid.depth2tiles(start - 1), disable=not cli_progress) as progress:
        for pos in pyramid.generate_pos(start):
            if pos.n == start:
                continue  # start layer is already there; we're cascading up

            # By construction, the children of this tile have all already been
            # processed.
            children = pyramid.pos_children(pos)

            img0 = pio.read_toasty_image(children[0], mode, default='none')
            img1 = pio.read_toasty_image(children[1], mode, default='none')
            img2 = pio.read_toasty_image(children[2], mode, default='none')
            img3 = pio.read_toasty_image(children[3], mode, default='none')

            if img0 is None and img1 is None and img2 is None and img3 is None:
                progress.update(1)
                continue  # No data here; ignore

            if buf is not None:
                buf.clear()

            for slidx, subimg in zip(SLICES, (img0, img1, img2, img3)):
                if subimg is not None:
                    if buf is None:
                        buf = mode.make_maskable_buffer(512, 512)

                    buf.asarray()[slidx] = subimg.asarray()

            merged = Image.from_array(mode, merger(buf.asarray()))
            pio.write_toasty_image(pos, merged)
            progress.update(1)

    if cli_progress:
        print()



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the AAS WorldWide Telescope project.
# Licensed under the MIT License.

"""Entrypoint for the "toasty" command-line interface.

"""
from __future__ import absolute_import, division, print_function

import argparse
import os.path
import sys


# General CLI utilities

def die(msg):
    print('error:', msg, file=sys.stderr)
    sys.exit(1)

def warn(msg):
    print('warning:', msg, file=sys.stderr)


# "cascade" subcommand

def cascade_getparser(parser):
    parser.add_argument(
        '--start',
        metavar = 'DEPTH',
        type = int,
        help = 'The depth of the TOAST layer to start the cascade',
    )
    parser.add_argument(
        'pyramid_dir',
        metavar = 'DIR',
        help = 'The directory containg the tile pyramid to cascade',
    )


def cascade_impl(settings):
    from .image import ImageMode
    from .merge import averaging_merger, cascade_images
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.pyramid_dir)

    start = settings.start
    if start is None:
        die('currently, you must specify the start layer with the --start option')

    cascade_images(pio, ImageMode.RGBA, start, averaging_merger, cli_progress=True)


# "healpix_sample_data_tiles" subcommand

def healpix_sample_data_tiles_getparser(parser):
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'fitspath',
        metavar = 'PATH',
        help = 'The HEALPix FITS file to be tiled',
    )
    parser.add_argument(
        'depth',
        metavar = 'DEPTH',
        type = int,
        help = 'The depth of the TOAST layer to sample',
    )


def healpix_sample_data_tiles_impl(settings):
    from .builder import Builder
    from .image import ImageMode
    from .pyramid import PyramidIO
    from .samplers import healpix_fits_file_sampler

    pio = PyramidIO(settings.outdir)
    sampler = healpix_fits_file_sampler(settings.fitspath)
    builder = Builder(pio)
    builder.toast_base(ImageMode.F32, sampler, settings.depth)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.fitspath}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "make_thumbnail" subcommand

def make_thumbnail_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        'imgpath',
        metavar = 'IN-PATH',
        help = 'The image file to be thumbnailed',
    )
    parser.add_argument(
        'outpath',
        metavar = 'OUT-PATH',
        help = 'The location of the new thumbnail file',
    )


def make_thumbnail_impl(settings):
    from .image import ImageLoader

    olp = settings.outpath.lower()
    if not (olp.endswith('.jpg') or olp.endswith('.jpeg')):
        warn('saving output in JPEG format even though filename is "{}"'.format(settings.outpath))

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    thumb = img.make_thumbnail_bitmap()

    with open(settings.outpath, 'wb') as f:
        thumb.save(f, format='JPEG')


# "multi_tan_make_data_tiles" subcommand

def multi_tan_make_data_tiles_getparser(parser):
    parser.add_argument(
        '--hdu-index',
        metavar = 'INDEX',
        type = int,
        default = 0,
        help = 'Which HDU to load in each input FITS file',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'paths',
        metavar = 'PATHS',
        nargs = '+',
        help = 'The FITS files with image data',
    )

def multi_tan_make_data_tiles_impl(settings):
    from .multi_tan import MultiTanDataSource
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.outdir)
    ds = MultiTanDataSource(settings.paths, hdu_index=settings.hdu_index)
    ds.compute_global_pixelization()

    print('Generating Numpy-formatted data tiles in directory {!r} ...'.format(settings.outdir))
    percentiles = ds.generate_deepest_layer_numpy(pio)

    if len(percentiles):
        print()
        print('Median percentiles in the data:')
        for p in sorted(percentiles.keys()):
            print('   {} = {}'.format(p, percentiles[p]))

    # TODO: this should populate and emit a stub index_rel.wtml file.


# "pipeline_fetch_inputs" subcommand

def _pipeline_add_io_args(parser):
    parser.add_argument(
        '--azure-conn-env',
        metavar = 'ENV-VAR-NAME',
        help = 'The name of an environment variable contain an Azure Storage '
                'connection string'
    )
    parser.add_argument(
        '--azure-container',
        metavar = 'CONTAINER-NAME',
        help = 'The name of a blob container in the Azure storage account'
    )
    parser.add_argument(
        '--azure-path-prefix',
        metavar = 'PATH-PREFIX',
        help = 'A slash-separated path prefix for blob I/O within the container'
    )
    parser.add_argument(
        '--local',
        metavar = 'PATH',
        help = 'Use the local-disk I/O backend'
    )

def _pipeline_io_from_settings(settings):
    from .pipeline import AzureBlobPipelineIo, LocalPipelineIo

    if settings.local:
        return LocalPipelineIo(settings.local)

    if settings.azure_conn_env:
        conn_str = os.environ.get(settings.azure_conn_env)
        if not conn_str:
            die('--azure-conn-env=%s provided, but that environment variable is unset'
                % settings.azure_conn_env)

        if not settings.azure_container:
            die('--azure-container-name must be provided if --azure-conn-env is')

        path_prefix = settings.azure_path_prefix
        if not path_prefix:
            path_prefix = ''

        return AzureBlobPipelineIo(
            conn_str,
            settings.azure_container,
            path_prefix
        )

    die('An I/O backend must be specified with the arguments --local or --azure-*')


def pipeline_fetch_inputs_getparser(parser):
    _pipeline_add_io_args(parser)
    parser.add_argument(
        'workdir',
        metavar = 'WORKDIR',
        default = '.',
        help = 'The local working directory',
    )

def pipeline_fetch_inputs_impl(settings):
    from .pipeline import PipelineManager

    pipeio = _pipeline_io_from_settings(settings)
    mgr = PipelineManager(pipeio, settings.workdir)
    mgr.fetch_inputs()


# "pipeline_process_todos" subcommand

def pipeline_process_todos_getparser(parser):
    _pipeline_add_io_args(parser)
    parser.add_argument(
        'workdir',
        metavar = 'WORKDIR',
        default = '.',
        help = 'The local working directory',
    )

def pipeline_process_todos_impl(settings):
    from .pipeline import PipelineManager

    pipeio = _pipeline_io_from_settings(settings)
    mgr = PipelineManager(pipeio, settings.workdir)
    mgr.process_todos()


# "pipeline_publish_todos" subcommand

def pipeline_publish_todos_getparser(parser):
    _pipeline_add_io_args(parser)
    parser.add_argument(
        'workdir',
        metavar = 'WORKDIR',
        default = '.',
        help = 'The local working directory',
    )

def pipeline_publish_todos_impl(settings):
    from .pipeline import PipelineManager

    pipeio = _pipeline_io_from_settings(settings)
    mgr = PipelineManager(pipeio, settings.workdir)
    mgr.publish_todos()


# "pipeline_reindex" subcommand

def pipeline_reindex_getparser(parser):
    _pipeline_add_io_args(parser)
    parser.add_argument(
        'workdir',
        metavar = 'WORKDIR',
        default = '.',
        help = 'The local working directory',
    )

def pipeline_reindex_impl(settings):
    from .pipeline import PipelineManager

    pipeio = _pipeline_io_from_settings(settings)
    mgr = PipelineManager(pipeio, settings.workdir)
    mgr.reindex()


# "tile_allsky" subcommand

def tile_allsky_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        '--placeholder-thumbnail',
        action = 'store_true',
        help = 'Do not attempt to thumbnail the input image -- saves memory for large inputs',
    )
    parser.add_argument(
        '--projection',
        metavar = 'PROJTYPE',
        default = 'plate-carree',
        help = 'The projection of the image; allowed choices are "plate-carree" and "plate-carree-planet"',
    )
    parser.add_argument(
        'imgpath',
        metavar = 'PATH',
        help = 'The image file to be tiled',
    )
    parser.add_argument(
        'depth',
        metavar = 'DEPTH',
        type = int,
        help = 'The depth of the TOAST layer to sample',
    )


def tile_allsky_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    pio = PyramidIO(settings.outdir)

    if settings.projection == 'plate-carree':
        from .samplers import plate_carree_sampler
        sampler = plate_carree_sampler(img.asarray())
    elif settings.projection == 'plate-carree-planet':
        from .samplers import plate_carree_planet_sampler
        sampler = plate_carree_planet_sampler(img.asarray())
    else:
        die('the image projection type {!r} is not recognized'.format(settings.projection))

    builder = Builder(pio)

    # Do the thumbnail first since for large inputs it can be the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    builder.toast_base(img.mode, sampler, settings.depth, cli_progress=True)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.imgpath}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "tile_study" subcommand

def tile_study_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        '--placeholder-thumbnail',
        action = 'store_true',
        help = 'Do not attempt to thumbnail the input image -- saves memory for large inputs',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'imgpath',
        metavar = 'PATH',
        help = 'The study image file to be tiled',
    )


def tile_study_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    img = ImageLoader.create_from_args(settings).load_path(settings.imgpath)
    pio = PyramidIO(settings.outdir)
    builder = Builder(pio)
    builder.default_tiled_study_astrometry()

    # Do the thumbnail first since for large inputs it can be the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    builder.tile_base_as_study(img, cli_progress=True)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.imgpath}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# "tile_wwtl" subcommand

def tile_wwtl_getparser(parser):
    from .image import ImageLoader
    ImageLoader.add_arguments(parser)

    parser.add_argument(
        '--placeholder-thumbnail',
        action = 'store_true',
        help = 'Do not attempt to thumbnail the input image -- saves memory for large inputs',
    )
    parser.add_argument(
        '--outdir',
        metavar = 'PATH',
        default = '.',
        help = 'The root directory of the output tile pyramid',
    )
    parser.add_argument(
        'wwtl_path',
        metavar = 'WWTL-PATH',
        help = 'The WWTL layer file to be processed',
    )


def tile_wwtl_impl(settings):
    from .builder import Builder
    from .image import ImageLoader
    from .pyramid import PyramidIO

    pio = PyramidIO(settings.outdir)
    builder = Builder(pio)
    img = builder.load_from_wwtl(settings, settings.wwtl_path)

    # Do the thumbnail first since for large inputs it can be the memory high-water mark!
    if settings.placeholder_thumbnail:
        builder.make_placeholder_thumbnail()
    else:
        builder.make_thumbnail_from_other(img)

    builder.tile_base_as_study(img, cli_progress=True)
    builder.write_index_rel_wtml()

    print(f'Successfully tiled input "{settings.wwtl_path}" at level {builder.imgset.tile_levels}.')
    print('To create parent tiles, consider running:')
    print()
    print(f'   toasty cascade --start {builder.imgset.tile_levels} {settings.outdir}')


# The CLI driver:

def entrypoint(args=None):
    """The entrypoint for the \"toasty\" command-line interface.

    Parameters
    ----------
    args : iterable of str, or None (the default)
      The arguments on the command line. The first argument should be
      a subcommand name or global option; there is no ``argv[0]``
      parameter.

    """
    # Set up the subcommands from globals()

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    commands = set()

    for py_name, value in globals().items():
        if py_name.endswith('_getparser'):
            cmd_name = py_name[:-10].replace('_', '-')
            subparser = subparsers.add_parser(cmd_name)
            value(subparser)
            commands.add(cmd_name)

    # What did we get?

    settings = parser.parse_args(args)

    if settings.subcommand is None:
        print('Run me with --help for help. Allowed subcommands are:')
        print()
        for cmd in sorted(commands):
            print('   ', cmd)
        return

    py_name = settings.subcommand.replace('-', '_')

    impl = globals().get(py_name + '_impl')
    if impl is None:
        die('no such subcommand "{}"'.format(settings.subcommand))

    # OK to go!

    impl(settings)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy as np
from numpy import testing as nt
import os.path
import pytest
import sys

from wwt_data_formats.filecabinet import FileCabinetWriter

from . import assert_xml_elements_equal, test_path
from .. import cli
from .. import study


class TestStudy(object):
    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_basic_cli(self):
        # First, create a WWTL. NB, filenames should match the ID's in the XML
        # file.

        fw = FileCabinetWriter()

        with open(test_path('layercontainer.wwtxml'), 'rb') as f:
            b = f.read()

        fw.add_file_with_data('55cb0cce-c44a-4a44-a509-ea66fce643a5.wwtxml', b)

        with open(test_path('NGC253ALMA.jpg'), 'rb') as f:
            b = f.read()

        fw.add_file_with_data('55cb0cce-c44a-4a44-a509-ea66fce643a5\\7ecb6411-e4ee-4dfa-90ef-77d6f486c7d2.jpg', b)

        with open(self.work_path('image.wwtl'), 'wb') as f:
            fw.emit(f)

        # Now run it through the CLI.

        for variants in ([], ['--placeholder-thumbnail']):
            args = ['tile-wwtl']
            args += variants
            args += [
                '--outdir', self.work_path('tiles'),
                self.work_path('image.wwtl')
            ]
            cli.entrypoint(args)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

import numpy as np
import numpy.testing as nt
import os.path
import pytest

from . import test_path
from .. import cli
from .. import merge


def test_averaging_merger():
    from ..merge import averaging_merger

    t = np.array([[np.nan, 1], [3, np.nan]])
    nt.assert_almost_equal(averaging_merger(t), [[2.]])


class TestCascade(object):
    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_basic_cli(self):
        """Test the CLI interface. We don't go out of our way to validate the
        computations in detail -- that's for the unit tests that probe the
        module directly.

        """
        for variants in ([], ['--placeholder-thumbnail']):
            args = ['tile-allsky']
            args += variants
            args += [
                '--outdir', self.work_path('basic_cli'),
                test_path('Equirectangular_projection_SW-tweaked.jpg'),
                '1',
            ]
            cli.entrypoint(args)

        args = [
            'cascade',
            '--start', '1',
            self.work_path('basic_cli'),
        ]
        cli.entrypoint(args)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy as np
from numpy import testing as nt
import os.path
import pytest
import shutil
import sys

from . import assert_xml_elements_equal, test_path
from .. import cli
from .. import pipeline


class LocalTestAstroPixCandidateInput(pipeline.AstroPixCandidateInput):
    def cache_data(self, cachedir):
        import json

        ext = 'jpg'
        shutil.copy(test_path('NGC253ALMA.jpg'), os.path.join(cachedir, 'image.' + ext))

        self._json['toasty_cached_image_name'] = 'image.' + ext

        with open(os.path.join(cachedir, 'astropix.json'), 'wt') as f:
            json.dump(self._json, f)


class LocalTestAstroPixImageSource(pipeline.AstroPixImageSource):
    def query_candidates(self):
        # NB all these values are wrong for the input image!!!
        item = {
            "creator": "Fake Observatory",
            "title": "Test",
            "description": "An amazing image.",
            "object_name": ["NGC 253"],
            "reference_url": "https://public.nrao.edu/gallery/astronomers-capture-first-image-of-a-black-hole/",
            "image_id": "test1",
            "image_credit": "Courtesy an amazing telescope.",
            "wcs_coordinate_frame": "ICRS",
            "wcs_equinox": "J2000",
            "wcs_reference_value": ["187.70593075", "12.39112325"],
            "wcs_reference_dimension": ["7416.0", "4320.0"],
            "wcs_reference_pixel": ["3738.9937831", "3032.00448074"],
            "wcs_scale": ["-5.91663506907e-14", "5.91663506907e-14"],
            "wcs_rotation": "0",
            "wcs_projection": "TAN",
            "wcs_quality": "Full",
            "wcs_notes": "FAKE",
            "publisher": "FAKE",
            "publisher_id": "fake",
            "resource_id": "test1",
            "last_updated": "2019-04-08T14:00:38.128143",
            "metadata_version": "1.1",
            "image_width": "7416",
            "image_height": "4320",
            "image_max_boundry": "7416",
            "astropix_id": 21642
        }
        yield LocalTestAstroPixCandidateInput(item)

    def open_input(self, unique_id, cachedir):
        import json

        with open(os.path.join(cachedir, 'astropix.json'), 'rt') as f:
            json_data = json.load(f)

        return pipeline.AstroPixInputImage(unique_id, cachedir, json_data)

pipeline._image_source_types['_local_test_astropix'] = LocalTestAstroPixImageSource


class TestPipeline(object):
    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

        os.makedirs(self.work_path('repo'))
        shutil.copy(test_path('toasty-pipeline-config.yaml'), self.work_path('repo'))

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_workflow(self):
        args = [
            'pipeline-fetch-inputs',
            '--local', self.work_path('repo'),
            self.work_path('work'),
        ]
        cli.entrypoint(args)

        args = [
            'pipeline-process-todos',
            '--local', self.work_path('repo'),
            self.work_path('work'),
        ]
        cli.entrypoint(args)

        args = [
            'pipeline-publish-todos',
            '--local', self.work_path('repo'),
            self.work_path('work'),
        ]
        cli.entrypoint(args)

        args = [
            'pipeline-reindex',
            '--local', self.work_path('repo'),
            self.work_path('work'),
        ]
        cli.entrypoint(args)

    def test_args(self):
        with pytest.raises(SystemExit):
            args = [
                'pipeline-fetch-inputs',
                self.work_path('work'),
            ]
            cli.entrypoint(args)

        with pytest.raises(SystemExit):
            args = [
                'pipeline-fetch-inputs',
                '--azure-conn-env', 'NOTAVARIABLE',
                self.work_path('work'),
            ]
            cli.entrypoint(args)

        os.environ['FAKECONNSTRING'] = 'fake'

        with pytest.raises(SystemExit):
            args = [
                'pipeline-fetch-inputs',
                '--azure-conn-env', 'FAKECONNSTRING',
                self.work_path('work'),
            ]
            cli.entrypoint(args)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

import numpy as np
import numpy.testing as nt
import os.path
import pytest

from . import test_path
from .. import cli
from .. import samplers


class TestSamplers(object):
    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_basic_cli(self):
        """Test some CLI interfaces. We don't go out of our way to validate the
        computations in detail -- that's for the unit tests that probe the
        module directly.

        """
        args = [
            'healpix-sample-data-tiles',
            '--outdir', self.work_path('basic_cli'),
            test_path('earth_healpix_equ.fits'),
            '1',
        ]
        cli.entrypoint(args)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Building up WWT imagery data sets.

This gets a little complex since the generation of a tiled image involves
several tasks that may or may not be implemented in several, swappable ways:
generating the tiled pixel data; positioning the image on the sky; filling in
metadata; and so on. We try to provide a framework that allows the
implementations of different tasks to be swapped out without getting too airy
and abstract.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
Builder
'''.split()

from wwt_data_formats.enums import DataSetType, ProjectionType
from wwt_data_formats.imageset import ImageSet
from wwt_data_formats.layers import ImageSetLayer, LayerContainerReader
from wwt_data_formats.place import Place

from .image import ImageLoader


class Builder(object):
    """
    State for some kind of imagery data set that's being assembled.
    """

    pio = None
    "A PyramidIO object representing the backing storage of the tiles and other image data."

    imgset = None
    """
    The WWT ImageSet data describing the image data and their positioning on the sky.

    Data URLs in this ImageSet should be populated as relative URLs.

    """
    place = None
    "The WWT Place data describing a default view of the image data."

    def __init__(self, pio):
        self.pio = pio
        self.imgset = ImageSet()
        self.imgset.name = 'Toasty'
        self.place = Place()
        self.place.foreground_image_set = self.imgset
        self.place.name = 'Toasty'


    def tile_base_as_study(self, image, **kwargs):
        from .study import tile_study_image

        tiling = tile_study_image(image, self.pio, **kwargs)
        tiling.apply_to_imageset(self.imgset)
        self.imgset.url = self.pio.get_path_scheme() + '.png'
        self.imgset.file_type = '.png'

        return self


    def default_tiled_study_astrometry(self):
        self.imgset.data_set_type = DataSetType.SKY
        self.imgset.base_degrees_per_tile = 1.0
        self.imgset.projection = ProjectionType.TAN
        self.place.zoom_level = 1.0
        return self


    def load_from_wwtl(self, cli_settings, wwtl_path):
        from io import BytesIO

        # Load WWTL and see if it matches expectations
        lc = LayerContainerReader.from_file(wwtl_path)

        if len(lc.layers) != 1:
            raise Exception('WWTL file must contain exactly one layer')

        layer = lc.layers[0]
        if not isinstance(layer, ImageSetLayer):
            raise Exception('WWTL file must contain an imageset layer')

        imgset = layer.image_set
        if imgset.projection != ProjectionType.SKY_IMAGE:
            raise Exception('WWTL imageset layer must have "SkyImage" projection type')

        # Looks OK. Read and parse the image.
        loader = ImageLoader.create_from_args(cli_settings)
        img_data = lc.read_layer_file(layer, layer.extension)
        img = loader.load_stream(BytesIO(img_data))

        # Transmogrify untiled image info to tiled image info. We reuse the
        # existing imageset as much as possible, but update the parameters that
        # change in the tiling process.
        self.imgset = imgset
        self.place.foreground_image_set = self.imgset
        wcs_keywords = self.imgset.wcs_headers_from_position()
        self.tile_base_as_study(img)
        self.imgset.set_position_from_wcs(wcs_keywords, img.width, img.height, place=self.place)

        return img


    def toast_base(self, mode, sampler, depth, **kwargs):
        from .toast import sample_layer
        sample_layer(self.pio, mode, sampler, depth, **kwargs)

        self.imgset.data_set_type = DataSetType.SKY
        self.imgset.base_degrees_per_tile = 180
        self.imgset.file_type = '.png'
        self.imgset.projection = ProjectionType.TOAST
        self.imgset.tile_levels = depth
        self.imgset.url = self.pio.get_path_scheme() + '.png'
        self.place.zoom_level = 360

        return self


    def cascade(self, **kwargs):
        from .image import ImageMode
        from .merge import averaging_merger, cascade_images
        cascade_images(self.pio, ImageMode.RGBA, self.imgset.tile_levels, averaging_merger, **kwargs)
        return self


    def make_thumbnail_from_other(self, thumbnail_image):
        thumb = thumbnail_image.make_thumbnail_bitmap()
        with self.pio.open_metadata_for_write('thumb.jpg') as f:
            thumb.save(f, format='JPEG')
        self.imgset.thumbnail_url = 'thumb.jpg'

        return self


    def make_placeholder_thumbnail(self):
        import numpy as np
        from .image import Image, ImageMode

        arr = np.zeros((45, 96, 3), dtype=np.uint8)
        img = Image.from_array(ImageMode.RGB, arr)

        with self.pio.open_metadata_for_write('thumb.jpg') as f:
            img.aspil().save(f, format='JPEG')

        self.imgset.thumbnail_url = 'thumb.jpg'
        return self


    def apply_wcs_info(self, wcs, width, height):
        self.imgset.set_position_from_wcs(
            wcs.to_header(),
            width, height,
            place = self.place,
        )

        return self


    def apply_avm_info(self, avm, width, height):
        self.apply_wcs_info(avm.to_wcs(target_shape=(width, height)), width, height)

        if avm.Title:
            self.imgset.name = avm.Title

        if avm.Description:
            self.imgset.description = avm.Description

        if avm.Credit:
            self.imgset.credits = avm.Credit

        if avm.ReferenceURL:
            self.imgset.credits_url = avm.ReferenceURL

        return self


    def write_index_rel_wtml(self):
        from wwt_data_formats import write_xml_doc
        from wwt_data_formats.folder import Folder

        self.place.name = self.imgset.name
        self.place.data_set_type = self.imgset.data_set_type
        self.place.thumbnail = self.imgset.thumbnail_url

        folder = Folder()
        folder.name = self.imgset.name
        folder.children = [self.place]

        with self.pio.open_metadata_for_write('index_rel.wtml') as f:
            write_xml_doc(folder.to_xml(), dest_stream=f, dest_wants_bytes=True)

        return self



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""Common routines for tiling images anchored to the sky in a gnomonic
(tangential) projection.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
StudyTiling
tile_study_image
'''.split()

import numpy as np
from tqdm import tqdm

from .pyramid import Pos, next_highest_power_of_2, tiles_at_depth


class StudyTiling(object):
    """
    Information about how a WWT "study" image is broken into tiles.

    In WWT a "study" is a large astronomical image projected onto the sky
    using a gnomonic (tangential or TAN) projection. The image may have many
    pixels, so it is broken up into tiles for efficient visualization.

    Note that this class doesn't know anything about how the image is
    projected onto the sky, or whether it is projected at all. The core tiling
    functionality doesn't need to care about that.

    """
    _width = None
    "The width of the region in which image data are available, in pixels (int)."

    _height = None
    "The height of the region in which image data are available, in pixels (int)."

    _p2n = None
    "The size of the study when tiled, in pixels - a power of 2."

    _tile_size = None
    "The number of tiles wide and high in the tiled study - a power of 2."

    _tile_levels = None
    "The number of levels in the tiling - ``log2(tile_size)``."

    _img_gx0 = None
    """The pixel position of the image data within the global (tiled)
    pixelization, in pixels right from the left edge (int). Nonnegative.

    """
    _img_gy0 = None
    """The pixel position of the image data within the global (tiled)
    pixelization, in pixels down from the top edge (int). Nonnegative.

    """
    def __init__(self, width, height):
        """Set up the tiling information.

        Parameters
        ----------
        width : positive integer
            The width of the full-resolution image, in pixels.
        height : positive integer
            The height of the full-resolution image, in pixels.

        """
        width = int(width)
        height = int(height)

        if width <= 0:
            raise ValueError('bad width value: %r' % (width, ))
        if height <= 0:
            raise ValueError('bad height value: %r' % (height, ))

        self._width = width
        self._height = height
        p2w = next_highest_power_of_2(self._width)
        p2h = next_highest_power_of_2(self._height)
        self._p2n = max(p2w, p2h)
        self._tile_size = self._p2n // 256
        self._tile_levels = int(np.log2(self._tile_size))
        self._img_gx0 = (self._p2n - self._width) // 2
        self._img_gy0 = (self._p2n - self._height) // 2


    def n_deepest_layer_tiles(self):
        """Return the number of tiles in the highest-resolution layer."""
        return 4**self._tile_levels


    def apply_to_imageset(self, imgset):
        """Fill the specific ``wwt_data_formats.imageset.ImageSet`` object
        with parameters defined by this tiling,

        Parameters
        ----------
        imgset : ``wwt_data_formats.imageset.ImageSet``
            The object to modify

        Notes
        -----
        The settings currently transferred are the number of tile levels and
        the projection type.

        """
        from wwt_data_formats.enums import ProjectionType

        imgset.tile_levels = self._tile_levels

        if self._tile_levels == 0:
          imgset.projection = ProjectionType.SKY_IMAGE
        else:
          imgset.projection = ProjectionType.TAN


    def image_to_tile(self, im_ix, im_iy):
        """Convert an image pixel position to a tiled pixel position.

        Parameters
        ----------
        im_ix : integer
            A 0-based horizontal pixel position in the image coordinate system.
        im_iy : integer
            A 0-based vertical pixel position in the image coordinate system.

        Notes
        -----
        ``(0, 0)`` is the top-left corner of the image. The input values need
        not lie on the image. (I.e., they may be negative.)

        Returns ``(tile_ix, tile_iy, subtile_ix, subtile_iy)``, where

        - *tile_ix* is X index of the matched tile in the tiling, between 0
          and 2**tile_size - 1. Measured right from the left edge of the
          tiling.

        - *tile_iy* is Y index of the matched tile in the tiling, between 0
          and 2**tile_size - 1. Measured down from the top of the tiling.

        - *subtile_ix* is the pixel X position within that tile, between 0 and
          255. Measured right from the left edge of the tiling.

        - *subtile_iy* is the pixel Y position within that tile, between 0 and
          255. Measured down from the top edge of the tiling.

        """
        gx = im_ix + self._img_gx0
        gy = im_iy + self._img_gy0
        tile_ix = np.floor(gx // 256).astype(np.int)
        tile_iy = np.floor(gy // 256).astype(np.int)
        return (tile_ix, tile_iy, gx % 256, gy % 256)


    def count_populated_positions(self):
        """
        Count how man tiles contain image data.

        This is used for progress reporting.

        """
        img_gx1 = self._img_gx0 + self._width - 1
        img_gy1 = self._img_gy0 + self._height - 1
        tile_start_tx = self._img_gx0 // 256
        tile_start_ty = self._img_gy0 // 256
        tile_end_tx = img_gx1 // 256
        tile_end_ty = img_gy1 // 256
        return (tile_end_ty + 1 - tile_start_ty) * (tile_end_tx + 1 - tile_start_tx)


    def generate_populated_positions(self):
        """Generate information about tiles containing image data.

        Generates a sequence of tuples ``(pos, width, height, image_x,
        image_y, tile_x, tile_y)`` where:

        - *pos* is a :class:`toasty.pyramid.Pos` tuple giving parameters of a tile
        - *width* is the width of the rectangle of image data contained in this tile,
          between 1 and 256.
        - *height* is the height of the rectangle of image data contained in this tile,
          between 1 and 256.
        - *image_x* is the pixel X coordinate of the left edge of the image data in this tile
          in the image rectangle, increasing from the left edge of the tile. Between 0 and
          ``self._width - 1`` (inclusive).
        - *image_y* is the pixel Y coordinate of the *top* edge of the image data in this tile
          in the image rectangle, increasing from the top edge of the tile. Between 0 and
          ``self._height - 1`` (inclusive).
        - *tile_x* is the pixel X coordinate of the left edge of the image data in this tile
          in the tile rectangle, increasing from the left edge of the tile. Between 0 and
          255 (inclusive).
        - *tile_y* is the pixel Y coordinate of the *top* edge of the image data in this tile
          in the tile rectangle, increasing from the top edge of the tile. Between 0 and
          255 (inclusive).

        Tiles that do not overlap the image at all are not generated. Tiles
        that are completely filled with image data will yield tuples of the
        form ``(pos, 256, 256, im_x, im_y, 0, 0)``. An image that fits
        entirely in one tile will yield a tuple of the form ``(Pos(n=0, x=0,
        y=0), width, height, 0, 0, tx, ty)``.

        """
        # Get the position of the actual image data in "global pixel
        # coordinates", which span the whole tiled region (a superset of the
        # image itself) with x=0, y=0 being the left-top corner of the tiled
        # region.

        img_gx1 = self._img_gx0 + self._width - 1  # inclusive: there are image data in this column
        img_gy1 = self._img_gy0 + self._height - 1  # ditto

        tile_start_tx = self._img_gx0 // 256
        tile_start_ty = self._img_gy0 // 256
        tile_end_tx = img_gx1 // 256  # inclusive; there are image data in this column of tiles
        tile_end_ty = img_gy1 // 256  # ditto

        for ity in range(tile_start_ty, tile_end_ty + 1):
            for itx in range(tile_start_tx, tile_end_tx + 1):
                # (inclusive) tile bounds in global pixel coords
                tile_gx0 = itx * 256
                tile_gy0 = ity * 256
                tile_gx1 = tile_gx0 + 255
                tile_gy1 = tile_gy0 + 255

                # overlap (= intersection) of the image and the tile in global pixel coords
                overlap_gx0 = max(tile_gx0, self._img_gx0)
                overlap_gy0 = max(tile_gy0, self._img_gy0)
                overlap_gx1 = min(tile_gx1, img_gx1)
                overlap_gy1 = min(tile_gy1, img_gy1)

                # coordinates of the overlap in image pixel coords
                img_overlap_x0 = overlap_gx0 - self._img_gx0
                img_overlap_x1 = overlap_gx1 - self._img_gx0
                img_overlap_y0 = overlap_gy0 - self._img_gy0
                img_overlap_y1 = overlap_gy1 - self._img_gy0

                # shape of the overlap
                overlap_width = img_overlap_x1 + 1 - img_overlap_x0
                overlap_height = img_overlap_y1 + 1 - img_overlap_y0

                # coordinates of the overlap in this tile's coordinates
                tile_overlap_x0 = overlap_gx0 - tile_gx0
                tile_overlap_y0 = overlap_gy0 - tile_gy0

                yield (
                    Pos(self._tile_levels, itx, ity),
                    overlap_width,
                    overlap_height,
                    img_overlap_x0,
                    img_overlap_y0,
                    tile_overlap_x0,
                    tile_overlap_y0,
                )


    def tile_image(self, image, pio, cli_progress=False):
        """Tile an in-memory image as a study.

        Parameters
        ----------
        image : :class:`toasty.image.Image`
            In-memory image data. The image's dimensions must match the ones
            for which this tiling was computed.
        pio : :class:`toasty.pyramid.PyramidIO`
            A handle for doing I/O on the tile pyramid
        cli_progress : optional boolean, defaults False
            If true, a progress bar will be printed to the terminal using tqdm.

        Returns
        -------
        Self.

        """
        if image.height != self._height:
            raise ValueError('height of image to be sampled does not match tiling')
        if image.width != self._width:
            raise ValueError('width of image to be sampled does not match tiling')

        buffer = image.mode.make_maskable_buffer(256, 256)

        with tqdm(total=self.count_populated_positions(), disable=not cli_progress) as progress:
            for pos, width, height, image_x, image_y, tile_x, tile_y in self.generate_populated_positions():
                iy_idx = slice(image_y, image_y + height)
                ix_idx = slice(image_x, image_x + width)
                by_idx = slice(tile_y, tile_y + height)
                bx_idx = slice(tile_x, tile_x + width)
                image.fill_into_maskable_buffer(buffer, iy_idx, ix_idx, by_idx, bx_idx)
                pio.write_toasty_image(pos, buffer)
                progress.update(1)

        if cli_progress:
            print()

        return self


def tile_study_image(image, pio, cli_progress=False):
    """Tile an image as a study, loading the whole thing into memory.

    Parameters
    ----------
    image : :class:`toasty.image.Image`
        The image to tile.
    pio : :class:`toasty.pyramid.PyramidIO`
        A handle for doing I/O on the tile pyramid
    cli_progress : optional boolean, defaults False
        If true, a progress bar will be printed to the terminal using tqdm.

    Returns
    -------
    A :class:`StudyTiling` defining the tiling of the image.

    """
    tiling = StudyTiling(image.width, image.height)
    tiling.tile_image(image, pio, cli_progress=cli_progress)
    return tiling
